#include <stdlib.h>
#include <string.h>
#include "list.h"

#include "ModuleLog.h"

#if(0) /* only needed when debugging */ 
static int List_IsMember(const struct list *pList, const struct list_node *pNode)
{
  const struct list_node *pSearch;

  REQUIRE(pList != NULL);
  REQUIRE(READABLE(pList));
  REQUIRE(pNode != NULL);
  REQUIRE(READABLE(pNode));

  pSearch = pList->pHead;
  if (pSearch == NULL) return 0;
  ASSERT(READABLE(pSearch));
  do
  {
    ASSERT(pSearch->sText != NULL);
    ASSERT(READABLE(pSearch->sText));
    ASSERT(pSearch->pNext != NULL);
    ASSERT(READABLE(pSearch->pNext));
    ASSERT(pSearch->pNext->pPrev == pSearch);

    if (pSearch == pNode)
      return 1;
    pSearch = pSearch->pNext;
  }
  while (pSearch != pList->pHead);

  return 0;
}

static int List_IsValid(const struct list *pList)
{
  const struct list_node *pNode;
  unsigned nEntries;

  if (!(pList != NULL)) return 0;
  if (!(READABLE(pList))) return 0;

  pNode = pList->pHead;
  if (pNode == NULL)
  {
    if (!(pList->pBrowse == NULL)) return 0;
    if (!(pList->nEntries == 0)) return 0;
  }
  else
  {
    if (!(pList->nEntries > 0)) return 0;
    if (!(pList->nMaxEntries == 0 || pList->nEntries < pList->nMaxEntries)) return 0;
    nEntries = pList->nEntries;
    if (!(READABLE(pNode))) return 0;
    do
    {
      if (!(pNode->pNext != NULL)) return 0;
      if (!(READABLE(pNode->pNext))) return 0;
      if (!(pNode->pNext->pPrev == pNode)) return 0;
      pNode = pNode->pNext;
    }
    while (--nEntries);
    if (!(pNode == pList->pHead)) return 0;
  }
  return 1;
}
#else
#define List_IsMember(LIST, NODE) 1
#define List_IsValid(LIST) 1
#endif

/***********************************************************
* struct list *List_Create()                               *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - unsigned nMaxEntries;                               *
*      Maximum length that the list will be allowed to     *
*      grow to.  0 means unbounded.                        *
*                         *  *  *  *                       *
* Creates a list with the specified limit.                 *
***********************************************************/
struct list *List_Create(unsigned nMaxEntries)
{
  struct list *pList = malloc(sizeof(*pList));
  if (pList)
  {
    pList->pBrowse =
    pList->pHead = NULL;
    pList->nEntries = 0;
    pList->nMaxEntries = nMaxEntries;
  }
  return pList;
}



/***********************************************************
* struct list *List_Destroy()                              *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - struct list *pList;                                 *
*      The list to destroy.                                *
*                         *  *  *  *                       *
* Frees a list and all of its nodes.                       *
***********************************************************/
struct list *List_Destroy(struct list *pList)
{
  if (pList)
  {
    if (pList->pHead)
    {
      /* break the loop */ 
      pList->pHead->pPrev->pNext = NULL;
      /* free nodes until we hit that break */ 
      while (pList->pHead)
      {
        pList->pBrowse = pList->pHead->pNext;
        free(pList->pHead->sText);
        free(pList->pHead);
        pList->pHead = pList->pBrowse;
      }
    }
    free(pList);
  }
  return NULL;
}


/***********************************************************
* static struct list_node *ListNode_Create()               *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - const char *sText;                                  *
*      The text for the new node.                          *
*                         *  *  *  *                       *
* Creates a new list node with a copy of the given text.   *
***********************************************************/
static struct list_node *ListNode_Create(const char *sText)
{
  struct list_node *pNode;
  unsigned cch;

  REQUIRE(sText != NULL);
  REQUIRE(READABLE(sText));

  pNode = malloc(sizeof(*pNode));
  if (pNode)
  {
    cch = strlen(sText);
    pNode->sText = malloc((cch+1) * sizeof(*pNode->sText));
    if (pNode->sText)
      strcpy(pNode->sText, sText);
    else
    {
      /* make failure atomic */ 
      free(pNode);
      pNode = NULL;
    }
  }

  ENSURE(pNode == NULL || WRITEABLE(pNode));
  ENSURE(pNode == NULL || READABLEN(pNode->sText, cch+1));
  ENSURE(pNode == NULL || strcmp(pNode->sText, sText) == 0);
  return pNode;
}


/***********************************************************
* static void ListNode_Link()                              *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - struct list_node *pNode;                            *
*      The node to link.                                   *
*                         *  *  *  *                       *
* Connects this node to the nodes listed in its pPrev and  *
* pNext fields.                                            *
***********************************************************/
static void ListNode_Link(struct list_node *pNode)
{
  REQUIRE(pNode != NULL);
  REQUIRE(READABLE(pNode));
  REQUIRE(pNode->pNext != NULL);
  REQUIRE(pNode->pPrev != NULL);
  REQUIRE(WRITEABLE(pNode->pNext));
  REQUIRE(WRITEABLE(pNode->pPrev));

  pNode->pNext->pPrev = pNode;
  pNode->pPrev->pNext = pNode;
}

/***********************************************************
* void ListNode_Unlink()                                   *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - struct list_node *pNode;                            *
*      The node to unlink.                                 *
*                         *  *  *  *                       *
* Disconnects the given node from those on either side.    *
***********************************************************/
static void ListNode_Unlink(struct list_node *pNode)
{
  REQUIRE(pNode != NULL);
  REQUIRE(READABLE(pNode));
  REQUIRE(pNode->pNext != NULL);
  REQUIRE(pNode->pPrev != NULL);
  REQUIRE(WRITEABLE(pNode->pNext));
  REQUIRE(WRITEABLE(pNode->pPrev));

  pNode->pNext->pPrev = pNode->pPrev;
  pNode->pPrev->pNext = pNode->pNext;
}

/***********************************************************
* struct list_node *List_InsertFirst()                     *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - struct list *pList;                                 *
*      The list.                                           *
*    - const char *sText;                                  *
*      The text for the node to insert.                    *
*                         *  *  *  *                       *
* Insert the first node in a list.  The list must be empty *
***********************************************************/
static struct list_node *List_InsertFirst(struct list *pList, const char *sText)
{
  struct list_node *pNode;

  REQUIRE(pList != NULL);
  REQUIRE(WRITEABLE(pList));
  REQUIRE(pList->pHead == NULL);
  REQUIRE(pList->nEntries == 0);

  pNode = ListNode_Create(sText);
  if (pNode)
  {
    pList->pHead = pNode;
    pNode->pNext = pNode->pPrev = pNode;
    ++pList->nEntries;
  }

  ENSURE(List_IsValid(pList));
  return pNode;
}

/***********************************************************
* struct list_node *List_InsertFront()                     *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - struct list *pList;                                 *
*      The list.                                           *
*    - const char *sText;                                  *
*      The text for the node to insert.                    *
*                         *  *  *  *                       *
* Inserts a new node at the front of the list.  The list   *
* must not be empty or at its limit.                       *
***********************************************************/
static struct list_node *List_InsertFront(struct list *pList, const char *sText)
{
  struct list_node *pNode;

  REQUIRE(pList != NULL);
  REQUIRE(WRITEABLE(pList));
  REQUIRE(pList->pHead != NULL);
  REQUIRE(pList->nEntries > 0);
  REQUIRE(pList->nMaxEntries == 0 || pList->nMaxEntries > pList->nEntries);

  pNode = ListNode_Create(sText);
  if (pNode)
  {
    pNode->pNext = pList->pHead;
    pNode->pPrev = pList->pHead->pPrev;
    ListNode_Link(pNode);
    pList->pHead = pNode;
    ++pList->nEntries;
  }

  ENSURE(List_IsValid(pList));
  return pNode;
}

/***********************************************************
* struct list_node *List_ReplaceBack()                     *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - struct list *pList;                                 *
*      The list.                                           *
*    - const char *sText;                                  *
*      The text for the node to insert.                    *
*                         *  *  *  *                       *
* Replaces the back node in the list and rotates it to the *
* front of the list.  The list should be at its limit.     *
***********************************************************/
static struct list_node *List_ReplaceBack(struct list *pList, const char *sText)
{
  unsigned cch;
  char *sNew;

  REQUIRE(pList != NULL);
  REQUIRE(WRITEABLE(pList));
  REQUIRE(pList->pHead != NULL);
  REQUIRE(pList->nEntries == pList->nMaxEntries);
  REQUIRE(sText != NULL);

  cch = strlen(sText);
  sNew = malloc((cch+1) * sizeof(*sNew));
  if (sNew)
  {
    strcpy(sNew, sText);
    pList->pHead = pList->pHead->pPrev;
    free(pList->pHead->sText);
    pList->pHead->sText = sNew;

    ENSURE(READABLEN(pList->pHead->sText, cch+1));
    ENSURE(strcmp(pList->pHead->sText, sText) == 0);

    ENSURE(List_IsValid(pList));
    return pList->pHead;
  }
  ENSURE(List_IsValid(pList));
  return NULL;
}

/***********************************************************
* struct list_node *List_MoveToFront()                     *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - struct list *pList;                                 *
*      The monkey.                                         *
*    - struct list_node *pNode;                            *
*      The node to move to the front of the list.  This    *
*      node must be a member of the list.                  *
*                         *  *  *  *                       *
* Moves the given node to the front of the list.           *
***********************************************************/
static struct list_node *List_MoveToFront(struct list *pList, struct list_node *pNode)
{
  REQUIRE(pList != NULL);
  REQUIRE(WRITEABLE(pList));
  REQUIRE(pNode != NULL);
  REQUIRE(WRITEABLE(pNode));
  REQUIRE(List_IsMember(pList, pNode));

  if (pNode != pList->pHead)
  {
    ListNode_Unlink(pNode);
    pNode->pNext = pList->pHead;
    pNode->pPrev = pList->pHead->pPrev;
    ListNode_Link(pNode);
    pList->pHead = pNode;
  }

  ENSURE(pList->pHead == pNode);
  ENSURE(List_IsValid(pList));
  return pNode;
}

/***********************************************************
* struct list_node *List_FindNode()                        *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - struct list *pList;                                 *
*      The list.                                           *
*    - const char *sText;                                  *
*      The text to search for.                             *
*                         *  *  *  *                       *
* Returns the first node matching the given text or NULL.  *
***********************************************************/
struct list_node *List_FindNode(struct list *pList, const char *sText)
{
  struct list_node *pNode;
  REQUIRE(sText != NULL);
  REQUIRE(READABLE(sText));

  if (pList && pList->pHead)
  {
    pNode = pList->pHead;
    do
    {
      ASSERT(pNode != NULL);
      ASSERT(READABLE(pNode));
      if (stricmp(sText, pNode->sText) == 0)
        return pNode;
      pNode = pNode->pNext;
    }
    while (pNode != pList->pHead);
  }
  return NULL;
}

/***********************************************************
* struct list_node *List_FindPrefixNode()                  *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - struct list *pList;                                 *
*      The list.                                           *
*    - const char *sPrefix;                                *
*      The prefix to search for.                           *
*    - unsigned cch;                                       *
*      The length of the prefix.                           *
*                         *  *  *  *                       *
* Finds the first node with a prefix matching that given.  *
***********************************************************/
struct list_node *List_FindPrefixNode(struct list *pList, const char *sPrefix, unsigned cch)
{
  /* It may be wise to find a faster way to do */ 
  /* this than a linear search, since it will  */ 
  /* be executed for each character entered in */ 
  /* the edit box.  Most people probably have  */ 
  /* short enough lists that it won't matter   */ 
  /* too much though.                          */ 

  struct list_node *pNode;
  REQUIRE(sPrefix != NULL);
  REQUIRE(READABLEN(sPrefix, cch));

  if (pList && pList->pHead)
  {
    pNode = pList->pHead;
    do
    {
      ASSERT(pNode != NULL);
      ASSERT(READABLE(pNode));
      ASSERT(pNode->sText != NULL);
      ASSERT(READABLE(pNode->sText));

      if (strnicmp(sPrefix, pNode->sText, cch) == 0)
        return pNode;
      pNode = pNode->pNext;
    }
    while (pNode != pList->pHead);
  }
  return NULL;
}

/***********************************************************
* void List_QueueInsert()                                  *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - struct list *pList;                                 *
*      The list.                                           *
*    - const char *sText;                                  *
*      The text to insert.                                 *
*                         *  *  *  *                       *
* Inserts a new node with the specified text, removing the *
* last node if the list has already reached its length     *
* limit.                                                   *
***********************************************************/
void List_QueueInsert(struct list *pList, const char *sText)
{
  REQUIRE(pList != NULL);
  REQUIRE(WRITEABLE(pList));
  REQUIRE(sText != NULL);
  REQUIRE(READABLE(sText));

  if (pList->nEntries == 0)
    List_InsertFirst(pList, sText);
  else if (pList->nMaxEntries == 0 || pList->nEntries < pList->nMaxEntries)
    List_InsertFront(pList, sText);
  else
    List_ReplaceBack(pList, sText);
}

/***********************************************************
* void List_QueueInsertUnique()                            *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - struct list *pList;                                 *
*      The list.                                           *
*    - const char *sText;                                  *
*      The text to insert.                                 *
*                         *  *  *  *                       *
* Inserts a new node with the specified text, removing     *
* duplicates, or the last node if the list has already     *
* reached its length limit and there are no duplicates.    *
***********************************************************/
void List_QueueInsertUnique(struct list *pList, const char *sText)
{
  struct list_node *pNode;
  REQUIRE(pList != NULL);
  REQUIRE(WRITEABLE(pList));
  REQUIRE(sText != NULL);
  REQUIRE(READABLE(sText));

  pNode = List_FindNode(pList, sText);
  if (pNode)
  {
    List_MoveToFront(pList, pNode);
    /* make sure it's the same case as the new version */ 
    ASSERT(READABLE(pNode));
    ASSERT(WRITEABLEN(pNode->sText, strlen(sText)+1));
    strcpy(pNode->sText, sText);
  }
  else
  {
    List_QueueInsert(pList, sText);
  }
}

/***********************************************************
* void List_Browse()                                       *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - struct list *pList;                                 *
*      The list to browse.                                 *
*    - int Direction;                                      *
*      One of the BROWSE_... constants.                    *
*    - int bWrap;                                          *
*      Nonzero if browsing off the end should wrap.        *
*    - unsigned nPageSize;                                 *
*      Number of elements to count as a page.              *
*                         *  *  *  *                       *
* Moves the browse pointer in a list.                      *
***********************************************************/
void List_Browse(struct list *pList, int Direction, int bWrap, unsigned nPageSize)
{
  REQUIRE(pList != NULL);
  REQUIRE(WRITEABLE(pList));

  if (pList->pHead == NULL)
    Direction = BROWSE_CANCEL;

  switch (Direction)
  {
  case BROWSE_CANCEL:
    pList->pBrowse = NULL;
    break;

  case BROWSE_HOME:
    pList->pBrowse = pList->pHead->pPrev;
    break;

  case BROWSE_END:
    pList->pBrowse = pList->pHead;
    break;

  case BROWSE_FORWARD:
    /* insertion always happens at the head, so pPrev is forwards */ 
    /* In retrospect, this was not a good desicion, as this fact  */ 
    /* will warp your brain if you try to make sense of the code. */ 
    if (!pList->pBrowse)
      pList->pBrowse = pList->pHead;
    else if (bWrap || pList->pBrowse != pList->pHead)
      pList->pBrowse = pList->pBrowse->pPrev;
    break;

  case BROWSE_BACK:
    if (!pList->pBrowse)
      pList->pBrowse = pList->pHead;
    else if (bWrap || pList->pBrowse != pList->pHead->pPrev)
      pList->pBrowse = pList->pBrowse->pNext;
    break;

  case BROWSE_PAGEFWD:
    REQUIRE(nPageSize != 0);
    /* move forward by a page, but stop if we hit    */ 
    /* the end in the middle.  If set to wrap, allow */ 
    /* pagefwd from the end after that.              */ 
    if (!bWrap && pList->pBrowse == pList->pHead)
      break;
    while (--nPageSize)
    {
      if (!pList->pBrowse)
        pList->pBrowse = pList->pHead;
      else
        pList->pBrowse = pList->pBrowse->pPrev;
      /* pause at the beginning */ 
      if (pList->pBrowse == pList->pHead)
        break;
    }
    /* is it the beginning, or is it the end??  */ 
    /* the end is the beginning, and the end is */ 
    /* right before the beginning.  Well, it    */ 
    /* works.  It works well.  I'm done with it */ 
    break;

  case BROWSE_PAGEBACK:
    REQUIRE(nPageSize != 0);
    /* move back by a page, but stop if we hit the   */ 
    /* beginning in the middle.  If set to wrap,     */ 
    /* allow pageback from the beginning after that. */ 
    if (!bWrap && pList->pBrowse == pList->pHead->pPrev)
      break;
    while (--nPageSize)
    {
      if (!pList->pBrowse)
        pList->pBrowse = pList->pHead;
      else
        pList->pBrowse = pList->pBrowse->pNext;
      /* pause at the end */ 
      if (pList->pBrowse == pList->pHead->pPrev)
        break;
    }
    break;

  default:
    TRACE1("List_Browse: invalid Direction");
  }
}
